# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Config/000_config.ipynb.

# %% auto 0
__all__ = ['logger', 'config_path_env', 'config_path_local', 'DB_HOSTS']

# %% ../../nbs/Config/000_config.ipynb 4
import logging
import os
import toml
from pathlib import Path

# Config.
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)

logger = logging.getLogger(__name__)

# Read local `config.toml` file.
config_path_env = os.getenv("BETTING_ENV_CONFIG")
config_path_local = Path(os.getenv("HOME")) / "secrets" / "config.toml"
config_path_local = config_path_local if config_path_local.exists() else Path("/secrets/config.toml")

if config_path_local.exists():
    CONFIG = toml.load(config_path_local)
elif not config_path_env is None:
    CONFIG = toml.load(config_path_env)
else:
    logger.error(
        "No config file found under `BETTING_ENV_CONFIG`, pkg root or in `/secrets/config.toml`!"
    )

# Defined database hosts.
DB_HOSTS = set([db for db in CONFIG["databases"]])



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/dataStrcuture/02_team_lineup.ipynb.

# %% auto 0
__all__ = ['Player', 'StartingPlayer', 'TeamSheet']

# %% ../../nbs/dataStrcuture/02_team_lineup.ipynb 3
import pandas as pd
import mongoengine
import datetime
import logging
from mongoengine.fields import ListField

# %% ../../nbs/dataStrcuture/02_team_lineup.ipynb 5
class Player(mongoengine.EmbeddedDocument):
    "Stores main information about a player reported on the team sheet"
    # Player Ids.
    opta_id = mongoengine.IntField(db_field="player_optaId")
    inStat_id = mongoengine.IntField(db_field="player_inStatId")
    fotMob_id = mongoengine.IntField(db_field="player_fotmobId")
    player_id = mongoengine.StringField(
        db_field="playerId", required=False, default="Unknown"
    )
    # Player name.
    name = mongoengine.StringField(
        db_field="playerName", required=False, default="Unknown"
    )
    # Player number.
    jersey_number = mongoengine.IntField(required=False)

    meta = {"allow_inheritance": True}


class StartingPlayer(Player):
    "Extends the `Player` class with extra information about a Player starting the game"
    # Formation id.
    formation_id = mongoengine.IntField(db_field="formationId", required=True)
    # Formation name.
    formation = mongoengine.StringField(db_field="formationName")
    # Formation slots.
    slot = mongoengine.IntField(required=True, choices=set(range(1, 12)))
    # Positions.
    position = mongoengine.StringField(required=True)
    # Lines.
    line = mongoengine.IntField(required=True, choices=set(range(1, 5)))

# %% ../../nbs/dataStrcuture/02_team_lineup.ipynb 7
class TeamSheet(mongoengine.Document):
    "Store the team-sheet information for a given game."
    
    # Team info.
    team_id = mongoengine.StringField(db_field="teamId", required=True)
    name = mongoengine.StringField(db_field="teamName")
    team_color = mongoengine.StringField(db_field="teamColor", required=False)

    # Optional metadata.
    is_official = mongoengine.BooleanField(
        db_field="is_official", required=False, default=True
    )
    competition_id = mongoengine.StringField(db_field="competitionId", required=False)
    competition_fotmob_id = mongoengine.StringField(
        db_field="competition_fotmobId", required=False
    )

    # Ids.
    opta_id = mongoengine.IntField(db_field="team_optaId")
    inStat_id = mongoengine.IntField(db_field="team_inStatId")
    fotMob_id = mongoengine.IntField(db_field="team_fotmobId")

    # Game ids.
    game_id = mongoengine.StringField(db_field="gameId", required=False)
    game_opta_id = mongoengine.IntField(db_field="game_optaId")
    game_inStat_id = mongoengine.IntField(db_field="game_inStatId")
    game_fotMob_id = mongoengine.IntField(db_field="game_fotmobId")
    game_date = mongoengine.DateTimeField(db_field="gameDate", required=False)

    # Manager info (optional).
    manager_id = mongoengine.StringField(db_field="managerId", required=False)
    manager_name = mongoengine.StringField(db_field="managerName", required=False)
    manager_fotmob_id = mongoengine.IntField(
        db_field="manager_fotmobId", required=False
    )

    # Referee name (optional).
    referee_name = mongoengine.StringField(db_field="refereeName", required=False)
    referee_country = mongoengine.StringField(db_field="refereeCountry", required=False)

    # Data-source.
    received_at = mongoengine.DateTimeField(
        db_field="timeReceived", default=datetime.datetime.now
    )
    source = mongoengine.StringField(db_field="lineupSource")

    # Players.
    starting = mongoengine.EmbeddedDocumentListField(StartingPlayer)
    bench = mongoengine.EmbeddedDocumentListField(Player)
    not_available = ListField(db_field="not_available", required=False)

    meta = {
        "db_alias": "features",
        "collection": "lineups",
        "ordering": ["-received_at"],
    }

    @classmethod
    def get_latest(
        cls,
        ra_team_id: str,  # Real-analytics game identifier.
        date: datetime.datetime = None, # Find the lastest data document prior to `date`.
    ): # TeamSheet output.
        if date is None:
            date = datetime.datetime.now()
        return (
            cls.objects(team_id=ra_team_id, received_at__lte=date, source="opta")
            .order_by("-received_at")
            .first()
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/dataStrcuture/01_game_features.ipynb.

# %% auto 0
__all__ = ['GameFeatures']

# %% ../../nbs/dataStrcuture/01_game_features.ipynb 3
import pandas as pd
import datetime
import mongoengine

# %% ../../nbs/dataStrcuture/01_game_features.ipynb 5
class GameFeatures(mongoengine.Document):
    "Store the game-features for a given game identified by its id"

    # Game-id.
    game_id = mongoengine.StringField(db_field="gameId", required=True)
    game_opta_id = mongoengine.IntField(db_field="game_optaId", required=False)

    # Game-date/time.
    game_date = mongoengine.DateTimeField(db_field="gameDate", required=True)
    season_id = mongoengine.IntField(db_field="seasonId", required=True)

    # Competition.
    competition_id = mongoengine.StringField(db_field="competitionId", required=True)

    # Team-ids.
    # Home.
    home_team_id = mongoengine.StringField(db_field="homeTeamId", required=True)
    ht_opta_id = mongoengine.IntField(db_field="homeTeam_optaId", required=False)

    # Away.
    away_team_id = mongoengine.StringField(db_field="awayTeamId", required=True)
    at_opta_id = mongoengine.IntField(db_field="awayTeam_optaId", required=False)

    # Stadium.
    venue_id = mongoengine.StringField(db_field="venueId", required=True)
    city = mongoengine.StringField(db_field="city", required=False)
    country = mongoengine.StringField(db_field="country", required=False)
    field_surface = mongoengine.StringField(db_field="surface", required=True)
    field_length_m = mongoengine.IntField(db_field="field_length_m", required=False)
    field_width_m = mongoengine.IntField(db_field="field_width_m", required=False)
    venue_capacity = mongoengine.IntField(db_field="capacity", required=False)
    venue_latitude = mongoengine.FloatField(db_field="lat", required=False)
    venuelongitude = mongoengine.FloatField(db_field="long", required=False)

    # Condition.
    is_closed_doors = mongoengine.BooleanField(
        db_field="is_behind_closed_doors", required=False
    )

    # Head to head.
    home_dominance = mongoengine.FloatField(
        db_field="homeTeamDominance", min_value=0, max_value=1, required=True
    )
    away_dominance = mongoengine.FloatField(
        db_field="awayTeamDominance", min_value=0, max_value=1, required=True
    )

    # Targets.
    tgt_half_gd = mongoengine.IntField(db_field="tgt_half_gd", required=False)
    tgt_gd = mongoengine.IntField(db_field="tgt_gd", required=False)
    tgt_htg = mongoengine.IntField(db_field="tgt_homeTeamGoals", required=False)
    tgt_atg = mongoengine.IntField(db_field="tgt_awayTeamGoals", required=False)
    tgt_half_htg = mongoengine.IntField(
        db_field="tgt_half_homeTeamGoals", required=False
    )
    tgt_half_atg = mongoengine.IntField(
        db_field="tgt_half_awayTeamGoals", required=False
    )
    tgt_outcome = mongoengine.FloatField(db_field="tgt_outcome", required=False)
    tgt_half_outcome = mongoengine.FloatField(
        db_field="tgt_half_outcome", required=False
    )

    meta = {
        "db_alias": "features",
        "collection": "gameFeatures_test",
    }

    @classmethod
    def get_all_games(
        cls,
        limit: int = None,  # Number of rows.
    ):  # GamesFeatures output.
        "Extract all games."
        return cls.objects().order_by("gameDate").limit(limit)

    @classmethod
    def get_game(
        cls,
        ra_game_id: str,  # Real-analytics game identifier.
    ):  # GamesFeatures output.
        "Extract a single game by its gameId."
        return cls.objects(game_id=ra_game_id).first()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/dataStrcuture/03_odds.ipynb.

# %% auto 0
__all__ = ['MARKET_TYPES', 'MarketOdds']

# %% ../../nbs/dataStrcuture/03_odds.ipynb 3
import datetime
import re
import mongoengine
import pandas as pd

# %% ../../nbs/dataStrcuture/03_odds.ipynb 5
MARKET_TYPES = ("asian", "total", "1x2")

# %% ../../nbs/dataStrcuture/03_odds.ipynb 6
class MarketOdds(mongoengine.Document):
    "Store the odds values"

    # Game-info.
    game_id = mongoengine.StringField(db_field="gameId", required=True)
    game_opta_id = mongoengine.IntField(db_field="game_optaId", required=False)
    game_date = mongoengine.DateTimeField(db_field="gameDate", required=False)
    competition_id = mongoengine.StringField(db_field="competitionId", required=True)
    competition_opta_id = mongoengine.IntField(
        db_field="competition_optaId", required=False
    )
    
    season_id = mongoengine.IntField(db_field="seasonId", required=True)
    home_team_id = mongoengine.StringField(db_field="homeTeamId", required=True)
    away_team_id = mongoengine.StringField(db_field="awayTeamId", required=True)

    # Bookmaker info.
    bookmaker_id = mongoengine.IntField(db_field="bookmakerId", required=True)
    bookmaker_name = mongoengine.StringField(db_field="bookmakerName", required=True)
    source = mongoengine.StringField(required=False)

    # Market info.
    market_type = mongoengine.StringField(
        choices=MARKET_TYPES, db_field="marketType", required=True
    )
    line_id = mongoengine.FloatField(db_field="lineId", required=False)

    # Odds.
    received_at = mongoengine.DateTimeField(db_field="timeReceived", required=False)
    odds1 = mongoengine.FloatField(required=True, min_value=1)
    odds2 = mongoengine.FloatField(required=True, min_value=1)
    oddsX = mongoengine.FloatField(required=False, min_value=1)
    overround = mongoengine.FloatField(required=False)
        
        
    meta = {
        "db_alias": "features",
        "collection": "odds",
        "ordering": ["-received_at"],
    }

    @classmethod
    def get_all_odds(
        cls,
        ra_game_id: str,  # Real-analytics game identifier.
        date: datetime.datetime = None,  # Find the lastest data document prior to `date`.
    ) -> pd.DataFrame:  # ALL markets information.
        "Extract latest odds (1x2, Asian handicap and Total) available for a given game prior to a given date"

        if date is None:
            date = datetime.datetime.now()

        odds_feats = cls.objects(game_id=ra_game_id, received_at__lt=date).order_by(
            "-received_at"
        )
        if not odds_feats:
            return None

        return pd.concat(
            [pd.DataFrame([{x: odds[x] for x in odds}]) for odds in odds_feats]
        ).reset_index(drop=True)

    @classmethod
    def get_latest_market(
        cls,
        ra_game_id: str,  # Real-analytics game identifier.
        market: str,  #  Type of market required; should one of MARKET_TYPES.
        date: datetime.datetime = None,  # Find the lastest data document prior to `date`.
    ):
        "Extract latest odds available for a given game prior to a given date"

        if date is None:
            date = datetime.datetime.now()

        if market == "asian":
            # Asian lines.
            odds_feats = cls.objects(
                game_id=ra_game_id, market_type=market, received_at__lt=date
            ).order_by("-received_at")

            # Case empty.
            if odds_feats is None:
                return None

            # Get all lines.
            all_lines = pd.DataFrame()
            for odds in odds_feats:
                odds_df = pd.DataFrame({x: odds[x] for x in odds}, index=[0])
                all_lines = pd.concat([all_lines, odds_df])

            # Remove duplicates and keep the fresh one.
            all_lines = all_lines.drop_duplicates(
                subset=["game_id", "line_id"], keep="first"
            )

            # Calculate delta between odds1 and odds 2.
            all_lines["delta"] = abs(all_lines["odds1"] - 2.0) + abs(
                all_lines["odds2"] - 2.0
            )
            # Keep the line that has a minimum delta (even line).
            line = all_lines.loc[all_lines["delta"] == all_lines["delta"].min()]

            return line

        else:
            return (
                cls.objects(
                    game_id=ra_game_id, market_type=market, received_at__lt=date
                )
                .order_by("-received_at")
                .first()
            )

    @classmethod
    def get_odds_features(
        cls,
        ra_game_id: str,  # Real-analytics game identifier.
        market: str = None,  # Type of market required; should one of MARKET_TYPES.
        date: datetime.datetime = None,  # Find the lastest data document prior to `date`.
    ) -> pd.DataFrame:  # 1-row Data.Frame with odds columns and probabilities columns.
        "Extract odds features and compute implied probs"

        if date is None:
            date = datetime.datetime.now()

        # Extract squad info.
        odds_feats = cls.get_latest_market(ra_game_id, market, date)
        if odds_feats is None:
            return None

        # Convert to dataFrame.
        odds_df = pd.DataFrame({x: odds_feats[x] for x in odds_feats}, index=[0])

        # Compute probs.
        odds_cols = odds_df.columns[odds_df.columns.str.startswith("odds")].tolist()
        probs_cols = [re.sub("odds", "proba", col) for col in odds_cols]
        odds_values = odds_df.loc[:, odds_cols]
        probs = pd.DataFrame(
            (1 / odds_values.values) / (1 / odds_values.values).sum(),
            index=[0],
            columns=probs_cols,
        )

        return pd.concat([odds_df, probs], axis=1)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/dataStrcuture/04_data_extractor.ipynb.

# %% auto 0
__all__ = ['data_aggregator']

# %% ../../nbs/dataStrcuture/04_data_extractor.ipynb 3
import datetime
import json
from typing import Tuple

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from ..config.mongo import mongo_init
from .game_features import *
from .odds import *
from .team_lineup import *

# %% ../../nbs/dataStrcuture/04_data_extractor.ipynb 6
def data_aggregator(
    limit: int = None,  # Number of rows to extract.
) -> pd.DataFrame:  # Mapped games.
    "Returns and aggregates games information from multiple Db collections."

    def _get_odds_columns(
        odds: pd.DataFrame,
    ) -> Tuple:  # set of odds
        "Returns Odds for a given market."
        if odds is None:
            return None, None, None, None
        
        if isinstance(odds, pd.DataFrame):
            odds = odds.head(1).squeeze()

        return odds.odds1, odds.odds2, odds.oddsX, odds.line_id

    def _keep_even_line(
        game_odds: pd.DataFrame,  # All markets for a given game.
    ) -> pd.DataFrame:
        "Returns indexes of the non-even lines."

        if game_odds is None:
            return None
        if not game_odds.shape[0]:
            return None

        # Calculate delta between "odds1" and "odds2" columns
        game_odds = game_odds.copy()
        game_odds["delta"] = abs(game_odds["odds1"] - 2.0) + abs(
            game_odds["odds2"] - 2.0
        )

        return game_odds.loc[game_odds.delta.idxmin(),].drop(columns="delta")

    def _odds(
        game_id: str,  # Real-analytics game identifier.
        game_date: datetime.datetime,  # Find the lastest data document prior to `date`.
    ) -> np.ndarray:  # Odds values.
        "Returns game Odds. It can be 1x2 or Asian Handicap."

        # Extract all odds(1x2, Asian handicap and Total).
        all_game_odds = MarketOdds.get_all_odds(ra_game_id=game_id, date=game_date)

        if all_game_odds is None:
            return None

        # select each market
        _odds_1x2 = all_game_odds[all_game_odds.market_type == "1x2"]
        
        _odds_asian = _keep_even_line(
            all_game_odds[all_game_odds.market_type == "asian"]
        )
        _odds_total = _keep_even_line(
            all_game_odds[all_game_odds.market_type == "total"]
        )

        # 1X2 odds.
        odds_1_1x2, odds_2_1x2, odds_x_1x2, _ = _get_odds_columns(_odds_1x2)

        # Asian Handicap.
        odds_1_ah, odds_2_ah, _, line_ah = _get_odds_columns(_odds_asian)

        # Total(Over/Under) odds.
        odds_1_total, odds_2_total, _, line_total = _get_odds_columns(_odds_total)

        return pd.DataFrame(
            {
                "preGameOdds1": odds_1_1x2,
                "preGameOddsX": odds_x_1x2,
                "preGameOdds2": odds_2_1x2,
                "preGameAhHome": odds_1_ah,
                "preGameAhAway": odds_2_ah,
                "preGameAhLineId": line_ah,
                "preGameOver": odds_1_total,
                "preGameUnder": odds_2_total,
                "preGameTotalLineId": line_total,
            },
            index=[0],
        )

    def _team_features(
        team_id: str,  # Real-analytics game identifier.
        game_date: datetime.datetime,  # Find the lastest data document prior to `date`.
    ) -> Tuple:  # Lineup values (name:position, ids, slots, formation name, timestamp).
        "Returns lineup features of a given team."

        # Lineup features.
        team_features = TeamSheet.get_latest(ra_team_id=team_id, date=game_date)

        if team_features is None:
            team_name = None
            team_lienups_ids = None
            team_lineups_names = None
            team_lienups_slots = None
            formation_name = None
            lineup_time_stamp = None
        else:
            # Team name.
            team_name = team_features.name
            # Players and positions.
            team_lineups_names = json.dumps(
                {player.name: player.position for player in team_features.starting}
            )
            # Players ids.
            team_lienups_ids = [player.opta_id for player in team_features.starting]
            # Players slots.
            team_lienups_slots = [player.slot for player in team_features.starting]
            # Formation name.
            formation_name = team_features.starting.first().formation
            # Lineup timestamp.
            lineup_time_stamp = team_features.received_at

        return pd.DataFrame(
            {
                "team_name": team_name,
                "team_lineups_names": [team_lineups_names],
                "team_lienups_ids": [team_lienups_ids],
                "team_lienups_slots": [team_lienups_slots],
                "formation_name": formation_name,
                "lineup_time_stamp": lineup_time_stamp,
            },
            index=[0],
        )
    
    

    # Extract games.
    games = GameFeatures.get_all_games(limit=limit)
    games = pd.DataFrame(games.as_pymongo())
    
    # Filter Data.
    games = games[
        [
            "gameId",
            "game_optaId",
            "gameDate",
            "homeTeamId",
            "homeTeam_optaId",
            "awayTeamId",
            "awayTeam_optaId",
            "tgt_gd",
            "tgt_outcome",
            "tgt_homeTeamGoals",
            "tgt_awayTeamGoals",
        ]
    ]

    # Map results {homewin -> 0 , draw -> 1, awaywin -> 2}.
    games["tgt_outcome"] = games["tgt_outcome"].map({1.0: 0.0, 0.0: 2.0, 0.5: 1.0})

    # compute other features
    def _one_game(row):
        o_1x2_ah_total = _odds(
            game_id=row["gameId"],
            game_date=row["gameDate"],
        )

        ht_feats = _team_features(
            team_id=row["homeTeamId"], game_date=row["gameDate"]
        ).rename(
            columns={
                "team_name": "homeTeamName",
                "team_lineups_names": "homeTeamLineup",
                "team_lienups_ids": "homeTeamLineupIds",
                "team_lienups_slots": "homeTeamLineupSlots",
                "formation_name": "homeTeamFormation",
                "lineup_time_stamp": "home_team_lineup_received_at",
            },
        )

        at_feats = _team_features(
            team_id=row["awayTeamId"], game_date=row["gameDate"]
        ).rename(
            columns={
                "team_name": "awayTeamName",
                "team_lineups_names": "awayTeamLineup",
                "team_lienups_ids": "awayTeamLineupIds",
                "team_lienups_slots": "awayTeamLineupSlots",
                "formation_name": "awayTeamFormation",
                "lineup_time_stamp": "away_team_lineup_received_at",
            },
        )

        res = pd.concat([o_1x2_ah_total, ht_feats, at_feats], axis=1)
        res.loc[:, "gameId"] = row.gameId

        return res

    return games.merge(
        pd.concat(
            [_one_game(row) for _, row in tqdm(games.iterrows(), total=games.shape[0])]
        ).reset_index(drop=True),
        on="gameId",
        how="left",
    )
